// -  

#include "/usr/src/sys/dev/ppbus/ppi.h"
#include "/usr/src/sys/dev/ppbus/ppbconf.h"
#include <unistd.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/resource.h>


// ߣ  ( /)
//#define k140
//#define k800


// FASTIO -  /   (" ").
//     ,     IO
//#define FASTIO

#ifdef FASTIO
#define READAC	3
#define WRITEAC	5
#else
#define READAC	2
#define WRITEAC	4
#endif

#define DRCHGAC	6

//      
int	fdPPI;

//    
#define	PPI_NAME	"/dev/ppi0"

//  
u_int8_t	io_Data, io_Status = 0xFF, io_Ctrl = 0x00;

// 
char state[128];


// - / 
void port_get() {
  ioctl(fdPPI, PPIGSTATUS, &io_Status);
};

void port_set() {
  ioctl(fdPPI, PPISDATA,   &io_Data);
  ioctl(fdPPI, PPISCTRL,   &io_Ctrl);
};


/*
	PC		AGAT
Data D7..D4	->	2 D6..D3
Status D6..D3	<-	0 D7..D4
Ctrl D0		=>inv	2 D7
Status D7	<=inv	1 D7
*/


//   tm    *rc
void io_05byte(unsigned char tm, unsigned char *rc) {
  strcpy(state, "05  A");
  do port_get(); while (! (io_Status & 0x80));	// ,    "0"   
  io_Data = tm << 4; port_set();		//  
  *rc = (io_Status >> 3) & 0x0F;		//  
  io_Ctrl |= 0x01; port_set();			//  "0"   
  strcpy(state, "05  B");
  do port_get(); while (  (io_Status & 0x80));	// ,    "1"   
  io_Ctrl &= 0xFE; port_set();			//  "1"   
//  printf(" %x -> , %x <-\n", tm, *rc);
};


//   tm    *rc
void io_byte(unsigned char tm, unsigned char *rc) {
  unsigned char l, h;

  io_05byte(tm & 0x0F, &l);
  io_05byte(tm   >> 4, &h);

  *rc = (h << 4) | l;
}


//   .   
void send_byte(unsigned char v) {
  unsigned char t;

  io_byte(v, &t);
}


//   .   0xFF
unsigned char recv_byte() {
  unsigned char t;

  io_byte(0xFF, &t);
  return t;  
}


//      ɣ
//  0,    ("1"   )
int data_present() {
  port_get();
  return io_Status & 0x80;
}


//  
void syncer() {
  unsigned char p;

  printf("Sync...\n");

loop:
  io_05byte(0x7, &p); if (p != 0x6) goto sync;
  io_05byte(0x5, &p); if (p != 0x4) goto sync;
  io_05byte(0x3, &p); if (p != 0x2) goto sync;
  io_05byte(0x1, &p); if (p != 0x0) goto sync;

  printf("Ok sync\n");
  return;

sync:
  printf("Fault. Wait for Enter...\n");
  getchar();
  printf("Retr...\n");
  goto loop;
}


//  
void open_ppi() {
  if ((fdPPI = open(PPI_NAME, O_RDWR)) == -1) { perror("open parallel"); exit(1); };
  ioctl(fdPPI, PPISSTATUS, &io_Status);
  ioctl(fdPPI, PPISCTRL,   &io_Ctrl);
}


//  
void close_ppi() {
  close(fdPPI);
}


//   
void reverse(unsigned char nsc, unsigned char *buf) {
  unsigned char tmp[(nsc+1)*256];
  int sc;

  for (sc = 0; sc <= nsc; sc++) {
    memmove(&(tmp[(nsc-sc)*256]), &(buf[sc*256]), 256);
  };
  memmove(buf, &tmp, sizeof(tmp));
}


//    
/*
 -:
0 -   ,  
1 -   
2 -  
3 -  
4 -  
5 -  
6 -  
       ,   
 RWTS    .
*/
//   -  0x45,  
void ctrl_out(unsigned char action, unsigned char tr, unsigned char nsc) {
  unsigned char d;

  do {
    syncer();
    printf(": %s, tr=%d, nsc=%d... ", (action & 0xFE) == (READAC & 0xFE) ? "" : "", tr, nsc);
    send_byte(action);
    send_byte(tr);
    send_byte(nsc);
    send_byte(action ^ tr ^ nsc);
    d = recv_byte();
    printf("CTRL_OUT: %x\n", d);
  } while (d != 0x45);
}


//  
void transmite_track(unsigned char tr, unsigned char nsc, unsigned char *buf) {
  int sc, ofs, tries, diff, err, eflag, fillcnt;
  unsigned char crc;
  unsigned char *p;

  tries = 0;
  do {
    if (tries++ > 2) { printf(" fault\n"); return; };
    p = buf; err = 0; fillcnt = 0;
    usleep(100000);
    ctrl_out(WRITEAC, tr, nsc);

    for (sc = 0; sc <= nsc; sc++) {
      diff = 0;
      for (ofs = 1; ofs < 256; ofs++) if (*(p + ofs) != *p) { diff = 1; break; };
      if ( diff ) {
        send_byte(0x83);
	crc = 0;
	for (ofs = 0; ofs < 256; ofs++) {
	  send_byte(*p);
	  crc ^= *p;
	  p++;
	};
        eflag = (crc != recv_byte());
	fillcnt++;
      } else {
        send_byte(0x27);
	send_byte(*p);
        crc = recv_byte();
	eflag = (crc != ((*p) ^ 0xFF));
	p += 256;
      };
      if (eflag) err++;
      putchar(eflag ? 'e' : '.');
    };

    printf("... %d error(s), %d/%d fill sectors.\n", err, fillcnt, nsc+1);
  } while (err != 0);
}


// ɣ 
void receive_track(unsigned char tr, unsigned char nsc, unsigned char *buf) {
  int sc, ofs, tries, diff, err, eflag, fillcnt;
  unsigned char crc;
  unsigned char *p;

  tries = 0;
  do {
retrace:
    if (tries++ > 2) { printf(" fault\n"); return; };
    p = buf; err = 0; fillcnt = 0;
    ctrl_out(READAC, tr, nsc);
    usleep(100000);
    
    for (sc = 0; sc <= nsc; sc++) {
      diff = recv_byte();
      if (diff == 0x27) {
        crc = recv_byte();
        for (ofs = 0; ofs < 256; ofs++) *p++ = crc;
	eflag = (crc != (recv_byte() ^ 0xFF));
      } else if (diff == 0x83) {
        crc = 0;
        for (ofs = 0; ofs < 256; ofs++) {
	  *p = recv_byte();
	  crc ^= *p;
	  p++;
	};
	eflag = (crc != recv_byte());
	fillcnt++;
//	printf("CRC: Calc: %x  Recv: %x\n", crc, rcrc);
      } else { goto retrace; };
      if (eflag) err++;
      putchar(eflag ? 'e' : '.');
    };

    printf("... %d error(s), %d/%d fill sectors.\n", err, fillcnt, nsc+1);
  } while (err != 0);
}


#ifdef k140
#define NTRACK	35
#define NSECTOR	16
#endif

#ifdef k800
#define NTRACK	160
#define NSECTOR	21
#endif


//  
void transmite_image(char *filename) {
  int		fd;
  unsigned char	tr, sc, buf[256*NSECTOR];

  if ((fd = open(filename, O_RDONLY)) == -1) { perror("open file"); exit(4); };
  for (tr = 0; tr < NTRACK; tr++) {
    if (read(fd, &buf, 256*NSECTOR) != 256*NSECTOR) { perror("read file"); exit(5); };
    reverse(NSECTOR-1, buf);
    transmite_track(tr, NSECTOR-1, buf);
  };
  close(fd);
}


// ɣ 
void receive_image(char *filename) {
  int		fd;
  unsigned char	tr, sc, buf[256*NSECTOR];

  if ((fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC)) == -1) { perror("open file"); exit(4); };
  for (tr = 0; tr < NTRACK; tr++) {
    receive_track(tr, NSECTOR-1, buf);
    reverse(NSECTOR-1, buf);
    write(fd, &buf, 256*NSECTOR);
  };    
  fchmod(fd, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  close(fd);
}


//     
void drive_change(char *param) {
  int s, d;

  s = atoi(param);
  d = (s % 10) & 1;
  s = s / 10;
  if (s > 6 || s < 1) s = 5;
  ctrl_out(DRCHGAC, s, d); 
  printf("Change drive to %d/%d\n", s, d);
}


main(int argc, char **argv) {
  int ch, i;
  unsigned char d;

  if (setpriority(PRIO_PROCESS, 0, 20)) perror("SetPriority: ");

  open_ppi();

  if (argc > 1) {
    while ((ch = getopt(argc, argv, "r:t:d:")) != -1) {
      switch (ch) {
        case 'r': receive_image(optarg); break;
        case 't': transmite_image(optarg); break;
	case 'd': drive_change(optarg); break;
      };
//      ctrl_out(0, 0, 0); //  
    };
  } else {
    printf("ARGS ? {{-r|-t} IMAGE.DSK | -d SD}\n");

/*
    for (i = 0; i < 256; i++) {
      d = recv_byte();
      printf(" : %02x", d);
    };

    for (i = 0; i < 256; i++) send_byte(i);
*/
  };

  close_ppi();
}
